// Generated file. DO NOT EDIT!
// aom_dsp needs a c file to force link language, 
// or to silence a harmless CMake warning: Ignore me.
void aom_aom_dsp_no_op_function(void);
void aom_aom_dsp_no_op_function(void) {}
